from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.dummytask.dummytask import DummyTask


class DummyTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(DummyTaskWidget, self).__init__(taskType=DummyTask, progressBar=progressBar)
    
    def validate(self) -> None:
        # self.showValidationError(parameterName='Some parameter', message='Something wrong')
        pass


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.totalsegmentatorroiselectiontask.totalsegmentatorroiselectiontask import TotalSegmentatorRoiSelectionTask

ROIS = [
    'adrenal_gland_left', 'adrenal_gland_right', 'aorta', 'autochthon_left', 'autochthon_right', 'brain', 'clavicula_left', 
    'clavicula_right', 'colon', 'duodenum', 'esophagus', 'face', 'femur_left', 'femur_right', 'gallbladder', 'gluteus_maximus_left', 
    'gluteus_maximus_right', 'gluteus_medius_left', 'gluteus_medius_right', 'gluteus_minimus_left', 'gluteus_minimus_right', 
    'heart_atrium_left', 'heart_atrium_right', 'heart_myocardium', 'heart_ventricle_left', 'heart_ventricle_right', 'hip_left', 
    'hip_right', 'humerus_left', 'humerus_right', 'iliac_artery_left', 'iliac_artery_right', 'iliac_vena_left', 'iliac_vena_right', 
    'iliopsoas_left', 'iliopsoas_right', 'inferior_vena_cava', 'kidney_left', 'kidney_right', 'liver', 'lung_lower_lobe_left', 
    'lung_lower_lobe_right', 'lung_middle_lobe_right', 'lung_upper_lobe_left', 'lung_upper_lobe_right', 'pancreas', 
    'portal_vein_and_splenic_vein', 'pulmonary_artery', 'rib_left_1', 'rib_left_2', 'rib_left_3', 'rib_left_4', 'rib_left_5', 
    'rib_left_6', 'rib_left_7', 'rib_left_8', 'rib_left_9', 'rib_left_10', 'rib_left_11', 'rib_left_12', 'rib_right_1', 'rib_right_2', 
    'rib_right_3', 'rib_right_4', 'rib_right_5', 'rib_right_6', 'rib_right_7', 'rib_right_8', 'rib_right_9', 'rib_right_10', 'rib_right_11', 
    'rib_right_12', 'sacrum', 'scapula_left', 'scapula_right', 'small_bowel', 'spleen', 'stomach', 'trachea', 'urinary_bladder', 
    'vertebrae_S1', 'vertebrae_C1', 'vertebrae_C2', 'vertebrae_C3', 'vertebrae_C4', 'vertebrae_C5', 'vertebrae_C6', 'vertebrae_C7', 
    'vertebrae_L1', 'vertebrae_L2', 'vertebrae_L3', 'vertebrae_L4', 'vertebrae_L5', 'vertebrae_T1', 'vertebrae_T2', 'vertebrae_T3', 
    'vertebrae_T4', 'vertebrae_T5', 'vertebrae_T6', 'vertebrae_T7', 'vertebrae_T8', 'vertebrae_T9', 'vertebrae_T10', 'vertebrae_T11', 
    'vertebrae_T12', 'heart', 'pulmonary_vein', 'brachiocephalic_trunk', 'subclavian_artery_right', 'subclavian_artery_left', 
    'common_carotid_artery_right', 'common_carotid_artery_left', 'brachiocephalic_vein_left', 'brachiocephalic_vein_right', 
    'atrial_appendage_left', 'superior_vena_cava', 'kidney_cyst_left', 'kidney_cyst_right', 'prostate', 'femur', 'patella', 
    'tibia', 'fibula', 'tarsal', 'metatarsal', 'phalanges_feet', 'humerus', 'ulna', 'radius', 'carpal', 'metacarpal', 'phalanges_hand', 
    'sternum', 'skull', 'subcutaneous_fat', 'skeletal_muscle', 'torso_fat', 'spinal_cord', 'lung_covid_infiltrate', 
    'intracerebral_hemorrhage', 'hip_implant', 'coronary_arteries', 'kidney', 'adrenal_gland', 'thyroid_gland', 'vertebrae_lumbar',
    'vertebrae_thoracic', 'vertebrae_cervical', 'iliac_artery', 'iliac_vena', 'ribs', 'scapula', 'clavicula', 'hip', 'gluteus_maximus', 
    'gluteus_medius', 'gluteus_minimus', 'autochthon', 'iliopsoas', 'lung_vessels', 'lung_trachea_bronchia', 'body_trunc', 
    'body_extremities', 'vertebrae_body', 'lung_pleural', 'pleural_effusion', 'pericardial_effusion', 'liver_vessels', 'liver_tumor', 
    'costal_cartilages'
]


class TotalSegmentatorRoiSelectionTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(TotalSegmentatorRoiSelectionTaskWidget, self).__init__(taskType=TotalSegmentatorRoiSelectionTask, progressBar=progressBar)
    
    def validate(self) -> None:
        pass


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.totalsegmentatorsliceselectiontask.totalsegmentatorsliceselectiontask import TotalSegmentatorSliceSelectionTask

ROIS = [
    'vertebrae_S1', 'vertebrae_C1', 'vertebrae_C2', 'vertebrae_C3', 'vertebrae_C4', 'vertebrae_C5', 'vertebrae_C6', 'vertebrae_C7', 
    'vertebrae_L1', 'vertebrae_L2', 'vertebrae_L3', 'vertebrae_L4', 'vertebrae_L5', 'vertebrae_T1', 'vertebrae_T2', 'vertebrae_T3', 
    'vertebrae_T4', 'vertebrae_T5', 'vertebrae_T6', 'vertebrae_T7', 'vertebrae_T8', 'vertebrae_T9', 'vertebrae_T10', 'vertebrae_T11', 
    'vertebrae_T12'
]


class TotalSegmentatorSliceSelectionTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(TotalSegmentatorSliceSelectionTaskWidget, self).__init__(taskType=TotalSegmentatorSliceSelectionTask, progressBar=progressBar)
    
    def validate(self) -> None:
        pass


import threading

from typing import Dict, List, Any

from PySide6.QtCore import QObject, Signal

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.descriptionparameter import DescriptionParameter
from mosamaticdesktop.tasks.labelparameter import LabelParameter
from mosamaticdesktop.tasks.filesetparameter import FileSetParameter
from mosamaticdesktop.tasks.multifilesetparameter import MultiFileSetParameter
from mosamaticdesktop.tasks.pathparameter import PathParameter
from mosamaticdesktop.tasks.filepathparameter import FilePathParameter
from mosamaticdesktop.tasks.textparameter import TextParameter
from mosamaticdesktop.tasks.integerparameter import IntegerParameter
from mosamaticdesktop.tasks.floatingpointparameter import FloatingPointParameter
from mosamaticdesktop.tasks.booleanparameter import BooleanParameter
from mosamaticdesktop.tasks.optiongroupparameter import OptionGroupParameter
from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.logger import Logger
from mosamaticdesktop.utils import createNameWithTimestamp

LOGGER = Logger()


class Task:
    class TaskProgressSignal(QObject):
        progress = Signal(int)
        finished = Signal(bool)

    IDLE = 0
    START = 1
    RUNNING = 2
    CANCELING = 3
    CANCELED = 4
    FINISHED = 5
    ERROR = 6

    @classmethod
    def NAME(cls):
        # Returns class name of child classes
        return cls.__qualname__

    def __init__(self) -> None:
        self._status = Task.IDLE
        self._progress = 0
        self._thread = None
        self._errors = []
        self._warnings = []
        self._info = []
        self._parameters = {}
        self._dataManager = DataManager()
        self._signal = self.TaskProgressSignal()

    def name(self) -> str:
        return self.__class__.__name__
    
    def logger(self) -> Logger:
        return LOGGER
    
    def parameters(self) -> List[Parameter]:
        return self._parameters.values()
    
    # def setParameters(self, parameters: Dict[str, Parameter]) -> None:
    #     self._parameters = parameters

    def updateParameters(self, parameters: Dict[str, Parameter]) -> None:
        for name in parameters.keys():
            if name in parameters.keys():
                self._parameters[name].setValue(parameters[name].value())
            else:
                LOGGER.warning(f'Task.updateParameters() parameter "{name}" does not exist')
    
    def parameter(self, name: str) -> Parameter:
        if name in self._parameters.keys():
            return self._parameters[name]
        return None
    
    def parameterValuesAsString(self) -> str:
        string = ''
        for name in self._parameters.keys():
            value = self._parameters[name].value()
            string += f'{name}="{value}", '
        return string[:-2]
    
    def addDescriptionParameter(self, name: str, description: str) -> Parameter:
        parameter = DescriptionParameter(name=name, description=description)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addLabelParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = LabelParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addFileSetParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FileSetParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter
    
    def addMultiFileSetParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = MultiFileSetParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addPathParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = PathParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addFilePathParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FilePathParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addTextParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = TextParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addIntegerParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, minimum: int=0, maximum: int=100, step: int=1) -> Parameter:
        parameter = IntegerParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, minimum=minimum, maximum=maximum, step=step)
        LOGGER.info(f'TaskWidget.addIntegerParameter() value={parameter.name()}')
        self._parameters[parameter.name()] = parameter
        return parameter

    def addFloatingPointParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FloatingPointParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter
    
    def addBooleanParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = BooleanParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addOptionGroupParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, options: List[str]=[]) -> Parameter:
        parameter = OptionGroupParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, options=options)
        self._parameters[parameter.name()] = parameter
        return parameter

    def signal(self):
        return self._signal
    
    def dataManager(self) -> DataManager:
        return self._dataManager
    
    # Logging
    
    def errors(self) -> List[str]:
        return self._errors
    
    def addError(self, message: str, toStdOut: bool=True, cancel=False) -> None:
        if toStdOut:
            LOGGER.error(message)
        self._errors.append(message)
        if cancel:
            self.cancel()

    def hasErrors(self) -> bool:
        return len(self._errors) > 0
    
    def warnings(self) -> List[str]:
        return self._warnings
    
    def addWarning(self, message: str, toStdOut: bool=True) -> None:
        if toStdOut:
            LOGGER.warning(message)
        self._warnings.append(message)

    def hasWarnings(self) -> bool:
        return len(self._warnings) > 0
    
    def info(self) -> List[str]:
        return self._info
    
    def addInfo(self, message: str, toStdOut: bool=True) -> None:
        if toStdOut:
            LOGGER.info(message)
        self._info.append(message)

    def hasInfo(self) -> bool:
        return len(self._info) > 0
    
    # Status
    
    def status(self) -> int:
        return self._status
    
    def statusIsIdle(self) -> bool:
        return self._status == Task.IDLE
    
    def statusIsStart(self) -> bool:
        return self._status == Task.START
    
    def statusIsRunning(self) -> bool:
        return self._status == Task.RUNNING
    
    def statusIsCanceling(self) -> bool:
        return self._status == Task.CANCELING
    
    def statusIsCanceled(self) -> bool:
        return self._status == Task.CANCELED
    
    def statusIsFinished(self) -> bool:
        return self._status == Task.FINISHED
    
    def statusIsError(self) -> bool:
        return self._status == Task.ERROR
    
    def setStatus(self, status: int) -> None:
        self._status = status

    def setStatusIdle(self) -> None:
        self._status = Task.IDLE

    def setStatusStart(self) -> None:
        self._status = Task.START

    def setStatusRunning(self) -> None:
        self._status = Task.RUNNING

    def setStatusCanceling(self) -> None:
        self._status = Task.CANCELING

    def setStatusCanceled(self) -> None:
        self._status = Task.CANCELED

    def setStatusFinished(self) -> None:
        self._status = Task.FINISHED

    def setStatusError(self) -> None:
        self._status = Task.ERROR

    # Progress

    def progress(self) -> int:
        return self._progress
    
    def updateProgress(self, step: int, nrSteps: int) -> None:
        if self._signal:
            self._progress = int(((step + 1) / (nrSteps)) * 100)
            self._signal.progress.emit(self._progress)
            if self._progress >= 100:
                self._signal.finished.emit(True)
                self._signal = None

    # Execution
                
    def start(self) -> None:
        self.setStatusStart()
        self._thread = threading.Thread(target=self.run)
        self._thread.start()
        self.setStatusRunning()

    def run(self) -> None:
        self.addInfo(f'Executing {self.name()}({self.parameterValuesAsString()})...')
        self.execute()
        if not self.statusIsCanceled():
            if self.hasErrors():
                self.setStatusError()
            else:
                self.setStatusFinished()        

    def execute(self) -> None:
        raise NotImplementedError()    

    def cancel(self) -> None:
        self.setStatusCanceled()
        self._thread.join()

    # Miscellaneous

    def generateTimestampForFileSetName(self, name: str) -> str:
        return createNameWithTimestamp(prefix=f'{name}')


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QComboBox, QVBoxLayout, QProgressBar

from mosamaticdesktop.widgets.dockwidget import DockWidget
from mosamaticdesktop.tasks.taskwidgetmanager import TaskWidgetManager
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TaskDockWidget(DockWidget):
    def __init__(self, title: str, progressBar: QProgressBar) -> None:
        super(TaskDockWidget, self).__init__(title)
        self._tasksComboBox = None
        self._placeholderWidget = None
        self._taskWidgetManager = TaskWidgetManager(progressBar=progressBar)
        self.initUi()

    def initUi(self) -> None:
        self._tasksComboBox = QComboBox(self)
        self._tasksComboBox.addItem(None)
        for taskName in sorted(self._taskWidgetManager.taskNames()):
            self._tasksComboBox.addItem(taskName)
        self._tasksComboBox.currentIndexChanged.connect(self.currentIndexChanged)
        self._placeholderWidget = QWidget()
        self._placeholderWidget.setLayout(QVBoxLayout())
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)
        layout.addWidget(self._tasksComboBox)
        layout.addWidget(self._placeholderWidget)
        widget = QWidget()
        widget.setLayout(layout)    
        self.setWidget(widget)
        self.setMinimumHeight(200)

    def currentIndexChanged(self, index) -> None:
        if self._placeholderWidget.layout():
            for i in reversed(range(self._placeholderWidget.layout().count())): 
                widgetToRemove = self._placeholderWidget.layout().itemAt(i).widget()
                self._placeholderWidget.layout().removeWidget(widgetToRemove)
                widgetToRemove.setParent(None)
        taskName = self._tasksComboBox.itemText(index)
        if taskName:
            taskWidget = self._taskWidgetManager.taskWidget(name=taskName)
            self._placeholderWidget.layout().addWidget(taskWidget)



import os

from PySide6.QtWidgets import QProgressBar

from typing import List, Dict

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.utils import ModuleLoader
from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


@singleton
class TaskWidgetManager:
    def __init__(self, progressBar: QProgressBar) -> None:
        self._progressBar = progressBar
        self._taskWidgetTypes = self.loadTaskWidgetTypes()
        self._taskTypes = self.loadTaskTypes()
        self._taskWidgets = {}
        self._tasks = {}

    def taskNames(self) -> List[str]:
        return list(self._taskTypes.keys())
    
    def taskWidgets(self) -> Dict[str, TaskWidget]:
        return self._taskWidgets
    
    def tasks(self) -> Dict[str, Task]:
        return self._tasks
    
    def taskWidget(self, name) -> TaskWidget:
        if name in self._taskWidgetTypes.keys():
            if name not in self._taskWidgets.keys():
                self._taskWidgets[name] = self._taskWidgetTypes[name](progressBar=self._progressBar)
            return self._taskWidgets[name]
        LOGGER.error(f'TaskWidgetManager: task widget for task {name} does not exist')
        return None
    
    def task(self, name) -> Task:
        if name in self._taskTypes.keys():
            if name not in self._tasks.keys():
                self._tasks[name] = self._taskTypes[name]()
            return self._tasks[name]
        LOGGER.error(f'TaskWidgetManager: task {name} does not exist')
        return None
    
    def loadTaskTypes(self):
        moduleDirectoryPath = os.path.dirname(os.path.realpath(__file__))
        LOGGER.info(f'TaskWidgetManager: loading tasks from {moduleDirectoryPath}...')
        taskWidgets = ModuleLoader.loadModuleClasses(
            moduleDirectoryPath=moduleDirectoryPath,
            moduleBaseClass=Task,
            fileNameEndsWith='task.py',
        )
        for taskName in taskWidgets.keys():
            LOGGER.info(f'TaskWidgetManager: Loaded task type {taskName}')
        return taskWidgets        

    def loadTaskWidgetTypes(self):
        moduleDirectoryPath = os.path.dirname(os.path.realpath(__file__))
        LOGGER.info(f'TaskWidgetManager: loading task widget types from {moduleDirectoryPath}...')
        taskWidgets = ModuleLoader.loadModuleClasses(
            moduleDirectoryPath=moduleDirectoryPath,
            moduleBaseClass=TaskWidget,
            fileNameEndsWith='taskwidget.py',
        )
        return taskWidgets


import os
import inspect

from mosamaticdesktop.utils import ParameterLoader
from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


@singleton
class ParameterCopier:
    """ This class is needed because when we pass parameters to the TaskWidgetParameterDialog
    they are deleted by the C++ backend after the dialog closes. Re-opening the dialog then
    results in errors. By explicitly making a copy of the parameters BEFORE passing them to
    the dialog, resolves this issue.
    """
    def __init__(self) -> None:
        self._parameterClasses = ParameterLoader.loadParameterClasses(
            parameterDirectoryPath=os.path.dirname(os.path.realpath(__file__)),
            parameterBaseClass=Parameter,
        )

    def makeCopy(self, obj):
        for cls in self._parameterClasses.values():
            # For some reason, isinstance(obj, cls) doesn't work...
            if obj.__class__.__name__ == cls.__name__:
                return obj.copy()


import os
import shutil

from totalsegmentator.python_api import totalsegmentator

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TotalSegmentatorRoiSelectionTask(Task):
    def __init__(self) -> None:
        super(TotalSegmentatorRoiSelectionTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Selects identical ROIs from TotalSegmentator outputs'
        )
        self.addPathParameter(
            name='rootDirectoryPath',
            labelText='Root Directory of TotalSegmentator Outputs'
        )
        self.addOptionGroupParameter(
            name='roi',
            labelText='ROI From TotalSegmentator Output',
            options=ROIS,
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path'
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get root directory path containing sub-directories for each CT scan
        rootDirectoryPath = self.parameter('rootDirectoryPath').value()
        if rootDirectoryPath:
            roi = self.parameter('roi').value()
            if roi:
                # Setup output fileset path
                outputFileSetName = self.parameter('outputFileSetName').value()
                if outputFileSetName is None:
                    outputFileSetName = self.generateTimestampForFileSetName(name=outputFileSetName)
                outputFileSetPath = self.parameter('outputFileSetPath').value()
                outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
                # Overwrite fileset
                overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
                self.addInfo(f'Overwrite output fileset: {overwriteOutputFileSet}')
                if overwriteOutputFileSet:
                    if os.path.isdir(outputFileSetPath):
                        shutil.rmtree(outputFileSetPath)
                os.makedirs(outputFileSetPath, exist_ok=True)
                self.addInfo(f'Output fileset path: {outputFileSetPath}')

                step = 0
                nrSteps = 0
                for scanDirectoryName in os.listdir(rootDirectoryPath):
                    scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                    if os.path.isdir(scanDirectoryPath):
                        nrSteps += 1

                for scanDirectoryName in os.listdir(rootDirectoryPath):
                    scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                    if os.path.isdir(scanDirectoryPath):

                        # Check if task was canceled first
                        if self.statusIsCanceled():
                            self.addInfo('Canceling task...')
                            break
                        
                        roiFilePath = os.path.join(scanDirectoryPath, roi + '.nii.gz')
                        roiFileName = scanDirectoryName + '-' + os.path.split(roiFilePath)[1]

                        self.addInfo(f'Copying {roiFilePath} to output fileset...')
                        shutil.copy(roiFilePath, os.path.join(outputFileSetPath, roiFileName))

                        self.updateProgress(step=step, nrSteps=nrSteps)
                        step += 1

                outputFileSet = self.dataManager().createFileSet(fileSetPath=outputFileSetPath)

                self.addInfo('Finished')
            else:
                self.addError('No ROI selected')
        else:
            self.addError('No root directory selected for TotalSegmentator outputs')


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.createpngfromsegmentationfiletask.createpngfromsegmentationfiletask import CreatePngFromSegmentationFileTask


class CreatePngFromSegmentationFileTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(CreatePngFromSegmentationFileTaskWidget, self).__init__(taskType=CreatePngFromSegmentationFileTask, progressBar=progressBar)
    
    def validate(self) -> None:
        pass


from typing import Dict

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QDialog, QPushButton, QVBoxLayout, QHBoxLayout, QMessageBox

from mosamaticdesktop.tasks.parameterwidget import ParameterWidget


class TaskWidgetParameterDialog(QDialog):
    def __init__(self, title: str, parametersWidgets: Dict[str, ParameterWidget], parent: QWidget=None) -> None:
        super(TaskWidgetParameterDialog, self).__init__(parent)
        self._title = title
        self._parametersWidgets = parametersWidgets
        self.initUi()

    def parameterWidgets(self) -> Dict[str, ParameterWidget]:
        return self._parametersWidgets

    def initUi(self) -> None:
        layout = QVBoxLayout()
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)
        layout.setAlignment(Qt.AlignTop)
        for name in self._parametersWidgets.keys():
            layout.addWidget(self._parametersWidgets[name])
        layout.addWidget(self.createButtonsWidget())
        self.setLayout(layout)
        self.setFixedWidth(500)
        self.setWindowTitle(self._title)

    def createButtonsWidget(self) -> None:
        cancelButton = QPushButton('Cancel')
        cancelButton.clicked.connect(self.cancel)
        saveAndCloseButton = QPushButton('Save and Close')
        saveAndCloseButton.setFocus()
        saveAndCloseButton.clicked.connect(self.saveAndClose)
        buttonsLayout = QHBoxLayout()
        buttonsLayout.addWidget(cancelButton)
        buttonsLayout.addWidget(saveAndCloseButton)
        buttonsLayout.setAlignment(Qt.AlignRight)
        buttonsWidget = QWidget()
        buttonsWidget.setLayout(buttonsLayout)
        return buttonsWidget
    
    def cancel(self) -> None:
        self.reject()

    def saveAndClose(self):
        for parameter in self._parametersWidgets.values():
            if not parameter.optional():
                if parameter.value() is None or parameter.value() == '':
                    QMessageBox.critical(self, 'Error', f'Parameter {parameter.name()} cannot be empty!')
                    return
        self.accept()

    def show(self):
        return self.exec_()



import sys

from PySide6.QtWidgets import QApplication

# from mainwindow import MainWindow
from mosamaticdesktop.mainwindow import MainWindow


def main():
    app = QApplication([])
    mainWindow = MainWindow()
    mainWindow.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()


import os
import ssl
import shutil
import whisper

# Requires installation of "ffmpeg" on your local system

from mosamaticdesktop.tasks.task import Task


class SpeechToTextConversionTask(Task):
    def __init__(self) -> None:
        super(SpeechToTextConversionTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Loads audio file (MP3) and converts it to text using Coqui-TTS package',
        )
        self.addFilePathParameter(
            name='inputFilePath',
            labelText='Input File Path to WAV Audio File',
        )
        self.addOptionGroupParameter(
            name='modelName',
            labelText='Model Name',
            options=['base', 'medium', 'large'],
            defaultValue='large',
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:
        # Hack: some CA certificate does not seem to be installed and giving errors when connecting to OpenAI
        ssl._create_default_https_context = ssl._create_unverified_context

        inputFilePath = self.parameter(name='inputFilePath').value()
        modelName = self.parameter(name='modelName').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if not outputFileSetName:
            outputFileSetName = self.generateTimestampForFileSetName(name='modelOutput')
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        step = 0
        nrSteps = 2
        self.addInfo(f'Loading model {modelName}...')
        model = whisper.load_model(modelName)
        self.updateProgress(step=step, nrSteps=nrSteps)
        step += 1
        self.addInfo(f'Transcribing audio file {inputFilePath}...')
        modelOutput = model.transcribe(inputFilePath)
        self.updateProgress(step=step, nrSteps=nrSteps)
        step += 1

        self.addInfo(f'Building output fileset: {outputFileSetPath}')
        with open(os.path.join(outputFileSetPath, 'modelOutput.txt'), 'w') as f:
            f.write(modelOutput['text'])
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        # Add warning to it automatically shows up with the transcribed text
        self.updateProgress(step=step, nrSteps=nrSteps)
        self.addError(modelOutput['text'])


import os
import shutil
import wave
import pyaudio

from mosamaticdesktop.tasks.task import Task


class SpeechRecordingTask(Task):
    def __init__(self) -> None:
        super(SpeechRecordingTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Records audio and saves it to .wav file',
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )
        self._frames = []

    def execute(self) -> None:
        outputFileSetName = self.parameter('outputFileSetName').value()
        if not outputFileSetName:
            outputFileSetName = self.generateTimestampForFileSetName(name='modelOutput')
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        p = pyaudio.PyAudio()
        stream = p.open(format=pyaudio.paInt16, channels=1, rate=44100, input=True, frames_per_buffer=1024)

        self.addInfo('Recording audio...')
        recording = True
        while recording:
            data = stream.read(1024)
            self._frames.append(data)
            if self.statusIsCanceled():
                self.addInfo('Stopping recording...')
                stream.stop_stream()
                stream.close()
                p.terminate()
                recording = False

        self.addInfo(f'Saving file recording.wav to: {outputFileSetPath}')
        wf = wave.open(os.path.join(outputFileSetPath, 'recording.wav'), 'wb')
        wf.setnchannels(1)
        wf.setsampwidth(pyaudio.PyAudio().get_sample_size(pyaudio.paInt16))
        wf.setframerate(44100)
        wf.writeframes(b''.join(self._frames))
        wf.close()        

        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        self.addInfo('Finished')


import os
import shutil
import threading

from PySide6.QtCore import Qt
from PySide6.QtGui import QWheelEvent #, Signal, QObject
from PySide6.QtWidgets import QWidget, QGraphicsView, QGraphicsScene, QProgressBar, QSlider, QCheckBox, QGridLayout, QLabel, QPushButton, QHBoxLayout
from PySide6.QtWidgets import QFileDialog

from mosamaticdesktop.widgets.viewers.dicomviewer.dicomlayer import DicomLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.dicominfolayer import DicomInfoLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.segmentationlayer import SegmentationLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.taglayer import TagLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.filesetcombobox import FileSetComboBox
from mosamaticdesktop.utils import isDicomFile
from mosamaticdesktop.data.fileset import FileSet
from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.data.file import File
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class DicomViewer(QWidget):
    # class ProgressSignal(QObject):
    #     progress = Signal(int)
    class MyGraphicsView(QGraphicsView):
        def __init__(self, parent=None):
            super(DicomViewer.MyGraphicsView, self).__init__(parent)

        def wheelEvent(self, event: QWheelEvent) -> None:
            event.ignore()

    def __init__(self, progressBar: QProgressBar) -> None:
        super(DicomViewer, self).__init__()
        self._progressBar = progressBar
        self._graphicsWidget = None
        self._graphicsView = None
        self._scene = None
        self._layerTuples = []
        self._windowLevel = 50
        self._windowWidth = 400
        # UI
        self._settingsWidget = None
        self._layerTupleSlider = None
        self._segmentationFileOpacityLabel = None
        self._segmentationFileOpacitySlider = None
        self._tagFileOpacityLabel = None
        self._tagFileOpacitySlider = None
        self._dicomFileInfoOpacityLabel = None
        self._dicomFileInfoOpacitySlider = None
        self._inputFileSetComboBox = None
        # Miscellaneous
        self._fullScan = False
        self._fullScanCheckBox = None
        self._currentLayerTupleIndex = 0
        self._exportDirectory = None
        self._dataManager = DataManager()
        self._layout = None
        # self._progressSignal = self.ProgressSignal()
        # self._progressSignal.progress.connect(self.updateProgress)
        self.initUi()

    def initUi(self) -> None:
        self.initGraphicsView()
        self.initSliders()
        self.initFileSetComboBox()
        self._fullScanCheckBox = QCheckBox('Full Scan')
        self._fullScanCheckBox.setCheckState(Qt.Checked if self._fullScan else Qt.Unchecked)
        self._fullScanCheckBox.stateChanged.connect(self.fullScanCheckBoxStateChanged)
        updateViewerButton = QPushButton('Update Viewer')
        updateViewerButton.clicked.connect(self.inputFileSetChanged)
        selectExportDirectoryButton = QPushButton('Select Export Directory')
        selectExportDirectoryButton.clicked.connect(self.selectExportDirectory)
        exportCurrentImageButton = QPushButton('Export Current Image to File')
        exportCurrentImageButton.clicked.connect(self.exportCurrentImage)
        layout = QGridLayout()
        layout.addWidget(self._graphicsView, 0, 0, 1, 2)
        layout.addWidget(QLabel('Image'), 1, 0)
        layout.addWidget(self._layerTupleSlider, 1, 1)
        layout.addWidget(QLabel('Opacity Segmentation'), 2, 0)
        layout.addWidget(self._segmentationFileOpacitySlider, 2, 1)
        layout.addWidget(QLabel('Opacity TAG'), 3, 0)
        layout.addWidget(self._tagFileOpacitySlider, 3, 1)
        layout.addWidget(QLabel('Input File Set'), 4, 0)
        layout.addWidget(self._inputFileSetComboBox, 4, 1)
        layout.addWidget(self._fullScanCheckBox, 5, 1)
        layout.addWidget(updateViewerButton, 6, 1)
        buttonLayout = QHBoxLayout()
        buttonLayout.addWidget(selectExportDirectoryButton)
        buttonLayout.addWidget(exportCurrentImageButton)
        layout.addLayout(buttonLayout, 7, 1)
        self.setLayout(layout)

    def initGraphicsView(self) -> None:
        self._graphicsView = DicomViewer.MyGraphicsView(self)
        self._scene = QGraphicsScene(self)
        item = self._scene.addText('DICOM Viewer')
        item.setDefaultTextColor(Qt.blue)
        self._graphicsView.setScene(self._scene)
        self._graphicsView.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._graphicsView.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    def initSliders(self) -> None:
        self._layerTupleSlider = QSlider(Qt.Horizontal, self)
        self._layerTupleSlider.setRange(0, 100)
        self._layerTupleSlider.valueChanged.connect(self.currentLayerTupleIndexChanged)
        self._segmentationFileOpacitySlider = QSlider(Qt.Horizontal, self)
        self._segmentationFileOpacitySlider.setRange(0, 100)
        self._segmentationFileOpacitySlider.setValue(50)
        self._segmentationFileOpacitySlider.valueChanged.connect(self.segmentationFileOpacityChanged)
        self._segmentationFileOpacitySlider.setVisible(False)
        self._tagFileOpacitySlider = QSlider(Qt.Horizontal, self)
        self._tagFileOpacitySlider.setRange(0, 100)
        self._tagFileOpacitySlider.setValue(50)
        self._tagFileOpacitySlider.valueChanged.connect(self.tagFileOpacityChanged)
        self._tagFileOpacitySlider.setVisible(False)

    def initFileSetComboBox(self) -> None:
        self._inputFileSetComboBox = FileSetComboBox(self)

    def findSegmentationFileForDicomFile(self, dicomFile: File, fileSet: FileSet) -> File:
        for file in fileSet.files():
            if file.name() == dicomFile.name() + '.seg.npy':
                return file
        return None

    def findTagFileForDicomFile(self, dicomFile: File, fileSet: FileSet) -> File:
        for file in fileSet.files():
            if file.name() == dicomFile.name() + '.tag':
                return file
            elif file.name() == dicomFile.name()[:-4] + '.tag':
                return file
            else:
                pass
        return None
    
    # def startSetInputFileSet(self, fileSet: FileSet):
    #     worker = threading.Thread(target=self.setInputFileSet, args=(fileSet, ))
    #     worker.start()

    def setInputFileSet(self, fileSet: FileSet) -> None:
        self._segmentationFileOpacitySlider.setVisible(False)
        self._tagFileOpacitySlider.setVisible(False)
        # self._progressBar.setValue(0)
        step = 0
        nrSteps = len(fileSet.files())
        for file in fileSet.files():
            layerTuple = [None, None, None, None] # dicom, numpy, tag, info
            if isDicomFile(filePath=file.path()):
                dicomLayer = DicomLayer()
                dicomLayer.setFile(file=file)
                dicomLayer.setWindowLevelAndWidth(windowLevel=self._windowLevel, windowWidth=self._windowWidth)
                layerTuple[0] = dicomLayer
                segmentationFile = self.findSegmentationFileForDicomFile(dicomFile=file, fileSet=fileSet)
                if segmentationFile:
                    segmentationLayer = SegmentationLayer()
                    segmentationLayer.setFile(file=segmentationFile)
                    layerTuple[1] = segmentationLayer
                    self._segmentationFileOpacitySlider.setVisible(True)
                tagFile = self.findTagFileForDicomFile(dicomFile=file, fileSet=fileSet)
                if tagFile:
                    tagLayer = TagLayer()
                    tagLayer.setFile(file=tagFile)
                    tagLayer.setShape([dicomLayer.data().Rows, dicomLayer.data().Columns])
                    layerTuple[2] = tagLayer
                    self._tagFileOpacitySlider.setVisible(True)
                dicomInfoLayer = DicomInfoLayer()
                dicomInfoLayer.setFile(file=file)
                dicomInfoLayer.setInstanceNumber(instanceNumber=dicomLayer.data().InstanceNumber)
                layerTuple[3] = dicomInfoLayer
                self._layerTuples.append(layerTuple)
            progress = int(((step + 1) / (nrSteps)) * 100)
            # QObject::setParent: Cannot set parent, new parent is in a different thread
            # QObject::startTimer: Timers cannot be started from another thread
            # self._progressSignal.progress.emit(progress)
            step += 1
        if self._fullScan:
            LOGGER.info('Sorting full scan images by instance number...')
            self._layerTuples = sorted(self._layerTuples, key=lambda layerTuple: layerTuple[0].data().InstanceNumber)
        self._layerTupleSlider.setRange(0, len(self._layerTuples) - 1)
        self.displayLayerTuple(self._currentLayerTupleIndex)

    # def updateProgress(self, progress: int) -> None:
    #     self._progressBar.setValue(progress)

    def currentLayerTupleIndexChanged(self, index) -> None:
        if index > 0 and index < len(self._layerTuples):
            self._currentLayerTupleIndex = index
            self.displayLayerTuple(self._currentLayerTupleIndex)

    def segmentationFileOpacityChanged(self, value) -> None:
        if self._layerTuples:
            for layerTuple in self._layerTuples:
                layerTuple[1].setOpacity(value / 100.0)
            self.displayLayerTuple(self._currentLayerTupleIndex)

    def tagFileOpacityChanged(self, value) -> None:
        if self._layerTuples:
            for layerTuple in self._layerTuples:
                layerTuple[2].setOpacity(value / 100.0)
            self.displayLayerTuple(self._currentLayerTupleIndex)

    def fullScanCheckBoxStateChanged(self, state) -> None:
        self._fullScan = True if state == Qt.Checked else False

    def inputFileSetChanged(self) -> None:
        text = self._inputFileSetComboBox.currentText()
        if text:
            self.clearData()
            inputFileSet = self._dataManager.fileSetByName(text)
            # self.startSetInputFileSet(fileSet=inputFileSet)
            self.setInputFileSet(fileSet=inputFileSet)

    def selectExportDirectory(self) -> None:
        self._exportDirectory = QFileDialog.getExistingDirectory(self, 'Select Export Directory')
        LOGGER.info(f'Selected export directory: {self._exportDirectory}')

    def exportCurrentImage(self) -> None:
        if self._currentLayerTupleIndex >= 0 and len(self._layerTuples) > 0:
            if self._exportDirectory:
                file = self._layerTuples[self._currentLayerTupleIndex][0].file()
                if file:
                    fileName = os.path.split(file.path())[1]
                    exportedFilePath = os.path.join(self._exportDirectory, fileName)
                    shutil.copy(file.path(), exportedFilePath)
                    LOGGER.info(f'Exported file {file.path()} to export directory')

    def wheelEvent(self, event) -> None:
        delta = event.angleDelta().y()
        if delta > 0 and self._currentLayerTupleIndex > 0:
            self._currentLayerTupleIndex -= 1
        elif delta < 0 and self._currentLayerTupleIndex < len(self._layerTuples) - 1:
            self._currentLayerTupleIndex += 1
        # Temporarily block slider events
        self._layerTupleSlider.valueChanged.disconnect(self.currentLayerTupleIndexChanged)
        self._layerTupleSlider.setValue(self._currentLayerTupleIndex)
        self._layerTupleSlider.valueChanged.connect(self.currentLayerTupleIndexChanged)
        # Display image
        self.displayLayerTuple(self._currentLayerTupleIndex)

    def displayLayerTuple(self, index: int) -> None:
        if len(self._layerTuples) > 0 and index < len(self._layerTuples):
            layerTuple = self._layerTuples[index]
            self._scene.clear()
            self._scene.addItem(layerTuple[0].createGraphicsItem())
            if layerTuple[1]:
                self._scene.addItem(layerTuple[1].createGraphicsItem())
            if layerTuple[2]:
                self._scene.addItem(layerTuple[2].createGraphicsItem())
            self._scene.addItem(layerTuple[3].createGraphicsItem())

    def clearData(self) -> None:
        self._scene.clear()
        self._layerTuples = []
        item = self._scene.addText('DICOM Viewer')
        item.setDefaultTextColor(Qt.blue)



import os
import json
import shutil
import numpy as np
import nibabel as nib

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.data.filecontentcache import FileContentCache
from mosamaticdesktop.utils import readFromCache, writeToCache
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class NumPyToNiftiConverterTask(Task):
    def __init__(self) -> None:
        super(NumPyToNiftiConverterTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Converts 2D or 3D NumPy arrays to NIFTI format'
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input File Set Name',
        )
        self.addTextParameter(
            name='transformationMatrix',
            labelText='Affine Transformation Matrix',
            optional=True,
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )
        self._cache = FileContentCache()

    def convertNumPyToNifti(self, numpyArray, transformationMatrix=None):
        if transformationMatrix is None:
            transformationMatrix = np.eye(4)
        if numpyArray.ndim == 2:
            numpyArray = numpyArray[:, :, np.newaxis]
        niftiImage = nib.Nifti1Image(numpyArray, transformationMatrix, dtype='int64')
        return niftiImage

    def execute(self) -> None:

        # Prepare input parameters
        inputFileSetName = self.parameter('inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(inputFileSetName)
        transformationMatrix = None
        transformationMatrixString = self.parameter('transformationMatrix').value()
        if transformationMatrixString and transformationMatrixString != '':
            try:
                transformationMatrix = json.loads(transformationMatrixString)
                transformationMatrix = np.array(transformationMatrix)
            except json.JSONDecodeError as e:
                self.addError(f'Could not load transformation matrix from parameter {transformationMatrixString} ({e})', cancel=True)

        # Prepare output parameters
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        step = 0
        files = inputFileSet.files()
        nrSteps = len(files)
        for file in files:

            # Chec if the task should cancel
            if self.statusIsCanceled():
                self.addInfo('Canceling task...')
                break

            if not file.name().endswith('.npy'):
                self.addWarning(f'Skipping non-NumPy file {file.name()}...')
                self.updateProgress(step=step, nrSteps=nrSteps)
                step += 1
                continue
            
            try:
                # Try to load file content from cache first. If it's not available
                # read it from disk
                content = readFromCache(file=file)
                if not content:
                    numpyArray = np.load(file.path())
                    content = writeToCache(file, numpyArray)
                numpyArray = content.fileObject()
                # Buid output file path
                outputFileName = file.name()[:-4] + '.nii.gz'
                outputFilePath = os.path.join(outputFileSetPath, outputFileName)
                # Convert NumPy array to NIFTI format and save
                niftiImage = self.convertNumPyToNifti(numpyArray=numpyArray, transformationMatrix=transformationMatrix)
                niftiImage.to_filename(outputFilePath)
            except OSError:
                self.addError(f'Could not load NumPy array from file: {file.path()}')

            # Update progress for this iteration         
            self.updateProgress(step=step, nrSteps=nrSteps)
            step += 1

        # Finalize task
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        self.addInfo('Finished')


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.createpngfromdicomfiletask.createpngfromdicomfiletask import CreatePngFromDicomFileTask


class CreatePngFromDicomFileTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(CreatePngFromDicomFileTaskWidget, self).__init__(taskType=CreatePngFromDicomFileTask, progressBar=progressBar)
    
    def validate(self) -> None:
        pass


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.speechtotextconversiontask.speechtotextconversiontask import SpeechToTextConversionTask


class SpeechToTextConversionTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(SpeechToTextConversionTaskWidget, self).__init__(taskType=SpeechToTextConversionTask, progressBar=progressBar)
    
    def validate(self) -> None:
        pass        


import inspect

from typing import Any, List, Dict

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QProgressBar, QProgressBar, QDialog, QMessageBox

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.tasks.taskwidgetexception import TaskWidgetException
from mosamaticdesktop.tasks.taskwidgetparameterdialog import TaskWidgetParameterDialog
from mosamaticdesktop.tasks.taskruninfodialog import TaskRunInfoDialog
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget
from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parametercopier import ParameterCopier
from mosamaticdesktop.tasks.descriptionparameter import DescriptionParameter
from mosamaticdesktop.tasks.descriptionparameterwidget import DescriptionParameterWidget
from mosamaticdesktop.tasks.labelparameter import LabelParameter
from mosamaticdesktop.tasks.labelparameterwidget import LabelParameterWidget
from mosamaticdesktop.tasks.filesetparameter import FileSetParameter
from mosamaticdesktop.tasks.filesetparameterwidget import FileSetParameterWidget
from mosamaticdesktop.tasks.multifilesetparameter import MultiFileSetParameter
from mosamaticdesktop.tasks.multifilesetparameterwidget import MultiFileSetParameterWidget
from mosamaticdesktop.tasks.pathparameter import PathParameter
from mosamaticdesktop.tasks.pathparameterwidget import PathParameterWidget
from mosamaticdesktop.tasks.filepathparameter import FilePathParameter
from mosamaticdesktop.tasks.filepathparameterwidget import FilePathParameterWidget
from mosamaticdesktop.tasks.textparameter import TextParameter
from mosamaticdesktop.tasks.textparameterwidget import TextParameterWidget
from mosamaticdesktop.tasks.integerparameter import IntegerParameter
from mosamaticdesktop.tasks.integerparameterwidget import IntegerParameterWidget
from mosamaticdesktop.tasks.floatingpointparameter import FloatingPointParameter
from mosamaticdesktop.tasks.floatingpointparameterwidget import FloatingPointParameterWidget
from mosamaticdesktop.tasks.booleanparameter import BooleanParameter
from mosamaticdesktop.tasks.booleanparameterwidget import BooleanParameterWidget
from mosamaticdesktop.tasks.optiongroupparameter import OptionGroupParameter
from mosamaticdesktop.tasks.optiongroupparameterwidget import OptionGroupParameterWidget
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TaskWidget(QWidget):    
    # Trick to return the child class name when NAME is retrieved
    # https://chat.openai.com/c/b7bd6334-5ec3-40e3-9af1-93405c68d795
    @classmethod
    def NAME(cls):
        # Returns class name of child classes (and strips of last 6 characters to get the task name itself)
        return cls.__qualname__[:-6]
    
    def __init__(self, taskType: Task, progressBar: QProgressBar) -> None:
        super(TaskWidget, self).__init__()
        if not inspect.isclass(taskType):
            raise TaskWidgetException('TaskWidget: argument taskType should be a class')
        self._taskType = taskType        
        self._progressBar = progressBar
        self._task = None
        self._taskParameterWidgets = self.createTaskParameterWidgetsFromTask(taskType=taskType)
        # We need this class to copy parameters to the TaskWidgetParameterDialog, otherwise the
        # parameters get deleted by C++ after the dialog closes.
        self._taskParameterCopier = ParameterCopier()
        ######
        # self._progressBar = None
        # self._progressBarLabel = None
        self._startButton = None
        self._cancelButton = None
        self._settingsButton = None
        self._runInfoButton = None
        self._test = False
        self.initUi()

    def name(self) -> str:
        return self.__class__.__name__
    
    def setTest(self, test: bool) -> None:
        self._test = test
    
    def initUi(self) -> None:
        self._startButton = QPushButton('Start')
        self._startButton.clicked.connect(self.startTask)
        self._cancelButton = QPushButton('Cancel')
        self._cancelButton.setEnabled(False)
        self._cancelButton.clicked.connect(self.cancelTask)
        self._settingsButton = QPushButton('Set Task Parameters...')
        self._settingsButton.clicked.connect(self.showTaskWidgetParameterDialog)
        self._runInfoButton = QPushButton('Run Info...')
        self._runInfoButton.setEnabled(False)
        self._runInfoButton.clicked.connect(self.showTaskRunInfoDialog)
        buttonLayout = QHBoxLayout()
        buttonLayout.addWidget(self._startButton)
        buttonLayout.addWidget(self._cancelButton)
        buttonLayout.addWidget(self._settingsButton)
        buttonLayout.addWidget(self._runInfoButton)
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)
        layout.addLayout(buttonLayout)
        self.setLayout(layout)

    # Task parameters
        
    def taskParameterWidgets(self) -> List[Parameter]:
        return self._taskParameterWidgets
        
    def taskParameterWidget(self, name: str) -> ParameterWidget:
        if name in self._taskParameterWidgets.keys():
            return self._taskParameterWidgets[name]
        return None
    
    def createTaskParameterWidgetsFromTask(self, taskType: Task) -> Dict[str, ParameterWidget]:
        widgets = {}
        task = taskType()
        for parameter in task.parameters():
            if isinstance(parameter, DescriptionParameter):
                widgets[parameter.name()] = DescriptionParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, LabelParameter):
                widgets[parameter.name()] = LabelParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, FileSetParameter):
                widgets[parameter.name()] = FileSetParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, MultiFileSetParameter):
                widgets[parameter.name()] = MultiFileSetParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, PathParameter):
                widgets[parameter.name()] = PathParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, FilePathParameter):
                widgets[parameter.name()] = FilePathParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, TextParameter):
                widgets[parameter.name()] = TextParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, IntegerParameter):
                widgets[parameter.name()] = IntegerParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, FloatingPointParameter):
                widgets[parameter.name()] = FloatingPointParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, BooleanParameter):
                widgets[parameter.name()] = BooleanParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, IntegerParameter):
                widgets[parameter.name()] = IntegerParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, OptionGroupParameter):
                widgets[parameter.name()] = OptionGroupParameterWidget(parameter=parameter, parent=self)
        return widgets
        
    # Task execution

    def startTask(self) -> None:
        self._task = self._taskType() # instantiate class
        # Map widgets back to parameters
        parameters = {}
        for taskParameterWidget in self._taskParameterWidgets.values():
            taskParameter = taskParameterWidget.parameter()
            parameters[taskParameter.name()] = taskParameter
        self._task.updateParameters(parameters=parameters)
        self._task.signal().progress.connect(self.taskProgress)
        self._task.signal().finished.connect(self.taskFinished)
        # Start task
        self._task.start()
        self._startButton.setEnabled(False)
        self._cancelButton.setEnabled(True)
        self._runInfoButton.setEnabled(False)
        self._progressBar.setValue(0)

    def cancelTask(self) -> None:
        if self._task:
            self._task.cancel()
            if not self._test:
                self._runInfoButton.setEnabled(True)
                self._cancelButton.setEnabled(False)
                self._startButton.setEnabled(True)
                self._progressBar.setValue(0)

    def showTaskWidgetParameterDialog(self) -> None:
        # Hack: we need to explicitly copy each parameter before passing them to
        # the TaskWidgetParameterDialog. If not, the Qt C++ backend will delete
        # the parameters if we try to show the dialog twice.
        parameterWidgets = {}
        for parameterWidget in self._taskParameterWidgets.values():
            parameterWidgets[parameterWidget.name()] = self._taskParameterCopier.makeCopy(parameterWidget)
        # Show task widget parameter dialog
        dialog = TaskWidgetParameterDialog(title=self.name(), parametersWidgets=self._taskParameterWidgets)
        result = dialog.show()
        if result == QDialog.Accepted:
            self._taskParameterWidgets = dialog.parameterWidgets()
            self.validate()

    def showTaskRunInfoDialog(self) -> None:
        if self._task:
            dialog = TaskRunInfoDialog(
                taskName=self._task.name(), errors=self._task.errors(), warnings=self._task.warnings(), info=self._task.info())
            dialog.show()
        
    def taskProgress(self, progress: int) -> None:
        if not self._test:
            self._progressBar.setValue(progress)

    def taskFinished(self, value: bool) -> None:
        if not self._test:
            self._runInfoButton.setEnabled(True)
            self._cancelButton.setEnabled(False)
            self._startButton.setEnabled(True)
            if self._task and (self._task.hasErrors() or self._task.hasWarnings()):
                self.showTaskRunInfoDialog()

    # Validation
        
    def validate(self) -> None:
        raise NotImplementedError()
    
    def showValidationError(self, parameterName: str, message: str) -> None:
        QMessageBox.critical(self, f'Validation Error for parameter "{parameterName}"', message)

    # Status (only used for testing)
        
    def taskIsIdle(self) -> bool:
        if self._task:
            return self._task.statusIsIdle()
        raise TaskWidgetException('TaskWidget: cannot determine task idle status, start it first')
        
    def taskIsStart(self) -> bool:
        if self._task:
            return self._task.statusIsStart()
        raise TaskWidgetException('TaskWidget: cannot determine task start status, start it first')
        
    def taskIsRunning(self) -> bool:
        if self._task:
            return self._task.statusIsRunning()
        raise TaskWidgetException('TaskWidget: cannot determine task running status, start it first')
        
    def taskIsCanceling(self) -> bool:
        if self._task:
            return self._task.statusIsCanceling()
        raise TaskWidgetException('TaskWidget: cannot determine task canceled status, start it first')
        
    def taskIsCanceled(self) -> bool:
        if self._task:
            return self._task.statusIsCanceled()
        raise TaskWidgetException('TaskWidget: cannot determine task canceled status, start it first')
        
    def taskIsFinished(self) -> bool:
        if self._task:
            return self._task.statusIsFinished()
        raise TaskWidgetException('TaskWidget: cannot determine task finished status, start it first')
    
    def taskIsError(self) -> bool:
        if self._task:
            return self._task.statusIsError()
        raise TaskWidgetException('TaskWidget: cannot determine task error status, start it first')


import os
import shutil

from totalsegmentator.python_api import totalsegmentator

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TotalSegmentatorTask(Task):
    def __init__(self) -> None:
        super(TotalSegmentatorTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Extracts anatomical ROIs from list of full CT scans'
        )
        self.addPathParameter(
            name='rootDirectoryPath',
            labelText='Root Directory of CT Scans (Each Scan as Separate Sub-Directory)'
        )
        self.addPathParameter(
            name='outputDirectoryPath',
            labelText='Output Directory with Segmentations'
        )
        self.addTextParameter(
            name='outputDirectoryName',
            labelText='Output Directory Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputDirectory',
            labelText='Overwrite Output Directory',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get root directory path containing sub-directories for each CT scan
        rootDirectoryPath = self.parameter('rootDirectoryPath').value()
        if rootDirectoryPath:
            outputDirectoryName = self.parameter('outputDirectoryName').value()
            if outputDirectoryName is None:
                outputDirectoryName = self.generateTimestampForFileSetName(name=outputDirectoryName)
            outputDirectoryPath = self.parameter('outputDirectoryPath').value()
            outputDirectoryPath = os.path.join(outputDirectoryPath, outputDirectoryName)
            self.addInfo(f'Output directory path: {outputDirectoryPath}')

            # Check whether to overwrite previous output directory
            overwriteOutputDirectory = self.parameter('overwriteOutputDirectory').value()
            self.addInfo(f'Overwrite output fileset: {overwriteOutputDirectory}')
            if overwriteOutputDirectory:
                if os.path.isdir(outputDirectoryPath):
                    shutil.rmtree(outputDirectoryPath)
            os.makedirs(outputDirectoryPath, exist_ok=True)

            # Each CT scan's files should be in a separate subdirectory inside the root directory
            step = 0
            nrSteps = 0
            for scanDirectoryName in os.listdir(rootDirectoryPath):
                scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                if os.path.isdir(scanDirectoryPath):
                    nrSteps += 1

            for scanDirectoryName in os.listdir(rootDirectoryPath):
                scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                if os.path.isdir(scanDirectoryPath):

                    # Check if task was canceled first
                    if self.statusIsCanceled():
                        self.addInfo('Canceling task...')
                        break

                    self.addInfo(f'Running TotalSegmentator on scan directory {scanDirectoryPath}...')
                    outputScanDirectoryPath = os.path.join(outputDirectoryPath, scanDirectoryName)
                    os.makedirs(outputScanDirectoryPath, exist_ok=True)
                    totalsegmentator(
                        scanDirectoryPath, outputScanDirectoryPath, fast=True)
                    
                    # Update progress for this iteration         
                    self.updateProgress(step=step, nrSteps=nrSteps)
                    step += 1

            self.addInfo('Finished processing scan directories')
            self.addInfo(f'Results can be found in: {outputDirectoryPath}')


import os
import shutil
import pydicom
import pydicom.errors
import nibabel as nib
import numpy as np

from typing import List, Dict, Union
from totalsegmentator.python_api import totalsegmentator

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.utils import currentTimeInSeconds, elapsedSeconds
from mosamaticdesktop.logger import Logger

LOGGER = Logger()

ROIS = [
    'vertebrae_S1', 'vertebrae_C1', 'vertebrae_C2', 'vertebrae_C3', 'vertebrae_C4', 'vertebrae_C5', 'vertebrae_C6', 'vertebrae_C7', 
    'vertebrae_L1', 'vertebrae_L2', 'vertebrae_L3', 'vertebrae_L4', 'vertebrae_L5', 'vertebrae_T1', 'vertebrae_T2', 'vertebrae_T3', 
    'vertebrae_T4', 'vertebrae_T5', 'vertebrae_T6', 'vertebrae_T7', 'vertebrae_T8', 'vertebrae_T9', 'vertebrae_T10', 'vertebrae_T11', 
    'vertebrae_T12'
]


class TotalSegmentatorSliceSelectionTask(Task):
    def __init__(self) -> None:
        super(TotalSegmentatorSliceSelectionTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Extracts vertebral images from CT scans'
        )
        self.addPathParameter(
            name='rootDirectoryPath',
            labelText='Root Directory of CT Scans (Each Scan as Separate Sub-Directory)'
        )
        self.addOptionGroupParameter(
            name='vertebra',
            labelText='Vertebral ROI From TotalSegmentator Output',
            options=ROIS,
        )
        self.addPathParameter(
            name='outputDirectoryPath',
            labelText='Output Directory with Selected Images'
        )
        self.addTextParameter(
            name='outputDirectoryName',
            labelText='Output Directory Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputDirectory',
            labelText='Overwrite Output Directory',
            defaultValue=True,
        )

    def loadNiftiFileAndAffineMatrix(self, path: str) -> Union[np.array, np.array]:
        niftiImage = nib.load(path)
        return niftiImage.get_fdata(), niftiImage.affine

    def calculateBoundingBoxInVoxelSpace(self, maskArray: np.array) -> List[int]:
        if 1 in np.unique(maskArray):
            indices = np.where(maskArray == 1)
            minX, maxX = np.min(indices[0]), np.max(indices[0])
            minY, maxY = np.min(indices[1]), np.max(indices[1])
            minZ, maxZ = np.min(indices[2]), np.max(indices[2])
            boundingBox = (minX, maxX, minY, maxY, minZ, maxZ)
            return boundingBox
        return None

    def calculateBoundingBoxInPatientOrientationSpace(self, boundingBox: List[int], affineMatrix: np.array) -> List[float]:
        min_x, max_x, min_y, max_y, min_z, max_z = boundingBox
        voxelCorners = np.array([
            [min_x, min_y, min_z, 1],
            [min_x, min_y, max_z, 1],
            [min_x, max_y, min_z, 1],
            [min_x, max_y, max_z, 1],
            [max_x, min_y, min_z, 1],
            [max_x, min_y, max_z, 1],
            [max_x, max_y, min_z, 1],
            [max_x, max_y, max_z, 1]
        ])
        transformedCorners = affineMatrix.dot(voxelCorners.T).T
        patientCoordinates = transformedCorners[:, :3]
        min_phys_x, max_phys_x = np.min(patientCoordinates[:, 0]), np.max(patientCoordinates[:, 0])
        min_phys_y, max_phys_y = np.min(patientCoordinates[:, 1]), np.max(patientCoordinates[:, 1])
        min_phys_z, max_phys_z = np.min(patientCoordinates[:, 2]), np.max(patientCoordinates[:, 2])
        boundingBoxInPatientOrientationSpace = (min_phys_x, max_phys_x, min_phys_y, max_phys_y, min_phys_z, max_phys_z)
        return boundingBoxInPatientOrientationSpace

    def calculateBoundingBoxes(self, path: str) -> Dict[str, np.array]:
        boundingBoxes = {}
        for niftiFileName in os.listdir(path):
            roiName = niftiFileName[:-7]
            niftiFilePath = os.path.join(path, niftiFileName)
            maskArray, affineMatrix = self.loadNiftiFileAndAffineMatrix(path=niftiFilePath)
            boundingBoxInVoxelSpace = self.calculateBoundingBoxInVoxelSpace(maskArray=maskArray)
            if boundingBoxInVoxelSpace:
                boundingBoxInPatientOrientationSpace = self.calculateBoundingBoxInPatientOrientationSpace(boundingBox=boundingBoxInVoxelSpace, affineMatrix=affineMatrix)
                self.addInfo(f'{roiName}: {boundingBoxInPatientOrientationSpace}')
                boundingBoxes[roiName] = boundingBoxInPatientOrientationSpace
            else:
                LOGGER.warning(f'{roiName}: empty mask')
        return boundingBoxes

    def execute(self) -> None:

        # Get root directory path containing sub-directories for each CT scan
        rootDirectoryPath = self.parameter('rootDirectoryPath').value()
        if rootDirectoryPath:
            outputDirectoryName = self.parameter('outputDirectoryName').value()
            if outputDirectoryName is None:
                outputDirectoryName = self.generateTimestampForFileSetName(name=outputDirectoryName)
            outputDirectoryPath = self.parameter('outputDirectoryPath').value()
            outputDirectoryPath = os.path.join(outputDirectoryPath, outputDirectoryName)
            self.addInfo(f'Output directory path: {outputDirectoryPath}')

            # Check whether to overwrite previous output directory
            overwriteOutputDirectory = self.parameter('overwriteOutputDirectory').value()
            self.addInfo(f'Overwrite output fileset: {overwriteOutputDirectory}')
            if overwriteOutputDirectory:
                if os.path.isdir(outputDirectoryPath):
                    shutil.rmtree(outputDirectoryPath)
            os.makedirs(outputDirectoryPath, exist_ok=True)

            # Get selected vertebra
            vertebra = self.parameter('vertebra').value()
            if vertebra:

                # Each CT scan's files should be in a separate subdirectory inside the root directory
                step = 0
                nrSteps = 0
                for scanDirectoryName in os.listdir(rootDirectoryPath):
                    scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                    if os.path.isdir(scanDirectoryPath):
                        nrSteps += 1

                for scanDirectoryName in os.listdir(rootDirectoryPath):
                    scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                    if os.path.isdir(scanDirectoryPath):

                        # Check if task was canceled first
                        if self.statusIsCanceled():
                            self.addInfo('Canceling task...')
                            break

                        # Store all images in dictionary by Z-coordinate
                        zCoords = {}
                        for fileName in os.listdir(scanDirectoryPath):
                            filePath = os.path.join(scanDirectoryPath, fileName)
                            try:
                                p = pydicom.dcmread(filePath, stop_before_pixels=True)
                                if 'ImagePositionPatient' in p:
                                    zCoords[p.ImagePositionPatient[2]] = filePath
                                else:
                                    self.addError(f'ImagePositionPatient attribute not in DICOM image')
                                    break
                            except pydicom.errors.InvalidDicomError:
                                self.addWarning(f'Image {filePath} of scan {scanDirectoryName} is not valid DICOM. Skipping...')
                                continue

                        self.addInfo(f'Running TotalSegmentator on scan directory {scanDirectoryPath}...')
                        outputScanDirectoryPath = os.path.join(outputDirectoryPath, scanDirectoryName)
                        os.makedirs(outputScanDirectoryPath, exist_ok=True)

                        # # Run TotalSegmentator to extract all vertebrae. We need this in order to check whether
                        # # the vertebra have the correct order and do not overlap
                        # start = currentTimeInSeconds()
                        # totalsegmentator(
                        #     scanDirectoryPath, 
                        #     outputScanDirectoryPath, 
                        #     fast=True,
                        #     roi_subset=ROIS,
                        # )
                        # self.addInfo(f'Elapsed: {elapsedSeconds(start)} seconds')

                        # https://github.com/MaastrichtU-CDS/2022_EvdWouwer_VertebraSegLabel/blob/main/TS_robustness_check.py
                        # Run error checks on extracted vertebrae. Start with the order of their Z-coordinates
                        boundingBoxes = self.calculateBoundingBoxes(path=outputScanDirectoryPath)
                        zMin = float('-inf')
                        for roi in ROIS:
                            if roi in boundingBoxes.keys():
                                boundingBox = boundingBoxes[roi]
                                self.addInfo(f'{roi}: zMin = {boundingBox[4]}')

                        # Select wanted vertebra

                        # Do slice selection on vertebra by taking middle slice
                        
                        # Update progress for this iteration         
                        self.updateProgress(step=step, nrSteps=nrSteps)
                        step += 1

                # Build output fileset

                self.addInfo('Finished processing scan directories')
                self.addInfo(f'Results can be found in: {outputDirectoryPath}')
            else:
                self.addError('No vertebra selected')
        else:
            self.addError('No root directory selected')


